# DIP2 CheatSheet

NumPy version: *1.19.5*
OpenCV version: *4.2.0*

## NumPy

### Axis fuckery
- Axis 0 runs downward the *rows* (parallel to the *columns*)
- Axis 1 runs alongside the *columns* (parallel to the *rows*)
- Axis 2 runs throug the *layers* (orthogonal to axis 0 and axis 1)

<img src="https://cdn-coiao.nitrocdn.com/CYHudqJZsSxQpAPzLkHFOkuzFKDpEHGF/assets/static/optimized/rev-b3f6ac8/wp-content/uploads/2018/12/numpy-arrays-have-axes_updated_v2.png" alt="Numpy axis explanation" width="400" />

Example:
```python
# Sum up values row-wise (alongside axis 1)
np.sum(array, axis=1)
```

### Creating arrays
- Create arbitrary numpy array
```python
w_gauss = np.array([[3, 6, 3], [6, 12, 6], [3, 6, 3]])
```
- Create numpy array filled with ones or zeroes
```python
zero_array = np.zeros((rows, cols), dtype = np.uint8)
ones_array = np.ones((rows, cols), dtype = np.uint64)
```

- Other usefull stuff
```python
np.arange(10,25,step=5) # evenly spaced values. Specify the stepsize
np.linspace(0,2,num=9, dtype) # evenly spaced values. Specify the number of values
np.eye(2, dtype) # 2x2 identity matrix
np.random.random((2,2)) # 2x2 array with random [0..1) values
np.random.randint(low, high, (2, 2)) # 2x2 array with random integers from [low, high[ (w/o high!)
```

### Array operations
```python
np.amax(array) # maximum of the whole array
np.amax(array, axis=1) #maxima along given axis (returns array)
np.amin(array) # minimum of the whole array
np.amin(array, axis=1) #minima along given axis (returns array)
np.maximum(array1, array2) # Returns new array with shape of array1 and array2 with element-wise maxima of each array
```

### Tricks
Calculate smallest *odd* interger not smaller than value (usefull for kernel size for gaussian filter)
```python
int(np.ceil(value) // 2 * 2 + 1)
```
Change type of the values of a numpy array. Possible types: *np.float32*,*np.float64*,*np.int32*, *np.[u]int16*, *np.[u]int32*, *np.[u]int64*
```python
array.astype(np.float64)
```

## OpenCV2

## Basic Stuff
- Load images with
```python
img_BGR = cv2.imread("path", cv2.IMREAD_COLOR)
img_GRAY = cv2.imread("path", cv2.IMREAD_GRAYSCALE)
```
- Colored images get loaded as BGR. Swap red and blue channel with
```python
RGB_img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
```
- In Grayscale images a 0 corresponds to black and 255 corresponds to white
- Obtain image shape, size and data type
```python
rows, cols (,channels) = img.shape #channels only if color image
n_pixels = img.size # = rows*cols*channels
img.dtype
```

## Image manipulation

### Resize
```python
cv2.resize(img, None, (0, 0), scale_x, scale_y, cv2.INTER_NEAREST)
cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT) #Pad with zeros. Top, bottom, etc. defines how many pixels get padded to each side
```

### Generate kernels
```python
gaus_kernel = cv2.getGaussianKernel(ksize, sigma)
(kx, ky) = cv2.getDerivKernels(order_x, order_y, ksize, normalize=False)
```

### Apply kernels
```python
# convolve image with kernel. ddepth=-1 keeps depth of input
cv2.filter2D(img, ddepth=-1, w_gauss/np.sum(w_gauss))
# when filtering can have negative outputs use a ddepth that can handle these!
cv2.filter2D(img, ddepth=cv2.CV_16S, w_sobel)
# Normalize image between given minimum and maximum
cv2.normalize(img, NONE, min_val, max_val, cv2.NORM_MINMAX)
```

### Frequency domain
```python
img_fft = np.fft.fft2(img) # compute fft
fft_shift = np.fft.fftshift(img_fft) # shift fft to the center
spec = np.absolute(fft_shift) # compute spectrum
#log_normalize before printing!
#open filter
filter = cv2.imread('filter.png', 0).astype(np.float64)
filter = filter / 255 # normalization
#apply filter
res_dft = np.mulitply(fft_shift, filter)
backshift = np.fft.ifftshift(res_dft) #backshift middle to the upper left
ifft = np.fft.ifft2(backshift) #inverse fft
res = np.real(ifft) # take only real-part
```

### log tranfsormation
```python
def log_transformation(img):
    # create empty image
    rows, cols = img.shape # image size
    res = np.zeros(shape=[rows, cols], dtype=np.float64)
    for i in range(rows):
        for j in range(cols):
            c = 255 / np.log(1 + np.amax(img)) # normalize result to 255*[0,1]
            res[i, j] = c * np.log(1 + img[i, j]) # min of log = 0, max = np.log(1 + img[i, j])
    # return result
    return res
```

## Plotting
### Setup
```python
from matplotlib import pyplot as plt
# set up plot
...
plt.title=("Title")
plt.xlabel=("x_label")
plt.ylabel=("y_label")
plt.show()
```

### Histogram
```python
array = img.flatten() #suitable for images (or 2D arrays in general)
plt.hist(array, num_bins, range)  #typical for grayscale: num_bins=255, range = [0, 256]
```

### Line
```python
plt.axvline(value, color="red") # vertical line at x="value"
plt.axhline(100, color="blue") # horizontal line at y=100
```

# Misc

## Depth combinations

|Input|Output|
|---|---|
|CV_8U|-1/CV_16S/CV_32F/CV_64F|
|CV_16U/CV_16S|-1/CV_32F/CV_64F|
|CV_32F|-1/CV_32F|
|CV_64F|-1/CV_64F|

- with ddepth=-1 output image will have same depth as input


# Exercises

List of the topics of each exercise to find them faster

## DIP 2
|Exercise|Topic|
|---|---|
|1|Filtering, Normalization, Histogram, Frequency Domain|
|2|Marr-Hildreth, Canny|
|3|Hough Transform, Region Growing|
|4|k-means Clustering, Simple Linear Iterative Clustering (SLIC, Superpixels)|
|5|Moore Boundary Following (Tracing), Fourier Descriptors|
|6|SIFT|
|7|Lucas-Kanade|
|8|Horn-Schnunck, Qualitative Optical Flow Evaluation|
|9|Normalized Correlation Coefficient (NCC) in 1D. Template Matching|
|10|Feature Extraction, Minimum Distance Classifier|
|11|Feature Extraction, Bayes Classifier|
|12|Least-Mean-Squared-Error (LMSE) Algorithm|
|13|Fully Connected Neural Networks and Backpropagation|
|14|Simple CNN to Classify Handwritten Numbers|

## DIP 1

|Exercise|Topic|
|---|---|
|1|Reading Images, Representation in 1D|
|2|Minimum/Maximum, Drawing Markers, Lines, Rectangles, Circles|
|3|Creating Periodic Patterns, Image Sampling, Creating Patterns with Variable Frequency, Moiré Effect|
|4|Thresholding, Negative Image, Gray-level Slicing, Range Scaling, Log Transformation, Gamma Correction, Bit-plane slicing|
|5|Histogram Computation and Analysis, Global/Local Histogram Equalization|
|6|Boundary Padding, Mean (Average) Filter, Arbitrary Filter Masks, Separable Filters|
|7|Gaussian Filter Masks, Ranking (Median) Filter, Differentiation of Images|
|8|Normalization and Clipping, Laplacian Filter, Laplacian of Gaussian (LoG), Roberts (Cross) and Sobel Operator|
|9|Discrete Fourier Transform, Inverse Discrete Fourier Transform (IDFT), Properties of the DFT and Spectrum, Spectrum and Phase Interpretation|
|10|Notch Filter, Ideal Lowpass Filter, Butterworth Lowpass Filter, Gaussian Lowpass Filter, Padding and Cropping|
|11|Highpass Filters, Image Restoration,Spatial Filters in the Frequency Domain, Template Matching (Correlation) in the Frequency Domain|
|12|Working with Color Images, RGB Color Space, HSV (HSI) Color Space, Intensity Slicing, Pseudo-color Enhancement|
|13|Converting Color Images to Grayscale Images, Intensity Adjustment, Color Complements, Color Slicing, Color Corrections, Color Segmentation in HSV|
|14|Run-Length Encoding (RLE), Bit-Plane Coding with Gray Code, Lempel-Ziv-Welch (LZW) Coding|
